"""Class to handle trajectory analysis of embeddings."""

import logging

import datashader as ds
import igraph as ig
import leidenalg
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import umap
import umap.plot
from datashader.mpl_ext import dsshow
from sklearn.cluster import KMeans
from sklearn.neighbors import kneighbors_graph
from sklearn.preprocessing import StandardScaler

from jointly_hic.core.config import TrajectoryAnalysisConfig

logger = logging.getLogger("joint_pca")


class TrajectoryAnalyzer:
    """Class to handle trajectory analysis of embeddings.

    TrajectoryAnalyzer takes embeddings and runs trajectory analysis on them.
    """

    def __init__(self, configuration: TrajectoryAnalysisConfig, embeddings: pd.DataFrame | None = None):
        """Initialize PostProcessor."""
        self.configuration: TrajectoryAnalysisConfig = configuration
        logger.info("TrajectoryAnalyzer :: configuration: \n\n%s", configuration)
        self.cluster_columns: list[str] = []

        # Pivot all PCs wide with chrom, start, end, good_bin as index and all PC* as columns
        if embeddings is None:
            logger.info(
                "TrajectoryAnalyzer :: Loading embeddings from parquet file: %s", self.configuration.parquet_file
            )
            loaded_embeddings: pd.DataFrame = pd.read_parquet(self.configuration.parquet_file)
        else:
            loaded_embeddings = embeddings

        trajectory_df = loaded_embeddings.pivot(
            index=["chrom", "start", "end"],
            columns="filename",
            values=[col for col in loaded_embeddings.columns if col.startswith(self.configuration.method)],
        )
        # Flatten column names to string
        column_names = trajectory_df.columns.to_flat_index()
        trajectory_df.columns = ["_".join(col) if type(col) is tuple else col for col in column_names]
        # Add good_bin column
        trajectory_df["good_bin"] = trajectory_df.notna().all(axis=1)
        # Add chrom, start, end columns
        trajectory_df.reset_index(inplace=True)
        logger.info("Shape of pivoted trajectory embeddings: %s", trajectory_df.shape)
        self.trajectory_df = trajectory_df

    @property
    def data_values(self) -> np.ndarray:
        """Return columns that start with 'PC#' and good_bin is True as numpy array."""
        return (
            self.trajectory_df.dropna()
            .loc[
                self.trajectory_df.good_bin,
                [col for col in self.trajectory_df.columns if col.startswith(self.configuration.method)],
            ]
            .values
        )

    def add_column(self, column_name, values):
        """Add column to embeddings where good_bin is True and otherwise add NaN."""
        self.trajectory_df[column_name] = np.nan
        self.trajectory_df.loc[self.trajectory_df.good_bin, column_name] = values

    def run(self):
        """Run post-processing."""
        logger.info("Running trajectory analysis")

        # Run KMeans clustering
        for n_clusters in self.configuration.kmeans_clusters:
            self.run_kmeans(self.data_values, n_clusters=n_clusters)

        # Run Leiden clustering
        self.run_leiden(self.data_values, n_neighbors=self.configuration.leiden_neighbors, preprocess=True)

        # Run UMAP
        for n_neighbors in self.configuration.umap_neighbours:
            self.run_umap(self.data_values, n_neighbors=n_neighbors)

        # Save trajectory_df to parquet and csv
        self.trajectory_df.to_parquet(f"{self.configuration.output}_trajectories.pq", index=False)
        self.trajectory_df.to_csv(f"{self.configuration.output}_trajectories.csv.gz", index=False)

        logger.info("Finished trajectory analysis")

    def run_umap(self, data: np.ndarray, n_neighbors: int = 50) -> None:
        """Run UMAP on embeddings."""
        logger.info("Running UMAP with %d neighbors", n_neighbors)
        mapper = umap.UMAP(n_neighbors=n_neighbors)
        umap_embedding = mapper.fit_transform(data)
        self.add_column(f"umap1_n{n_neighbors}", umap_embedding[:, 0])
        self.add_column(f"umap2_n{n_neighbors}", umap_embedding[:, 1])

        # Plot a umap for each clustering_column
        n_clusters = len(self.cluster_columns)
        grid_size = int(np.ceil(np.sqrt(n_clusters)))
        _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
        for i, cluster_column in enumerate(self.cluster_columns):
            plotdf = self.trajectory_df.loc[
                self.trajectory_df.good_bin, [f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"]
            ]
            plotdf[cluster_column] = self.trajectory_df.loc[self.trajectory_df.good_bin, cluster_column].astype(
                "category"
            )
            # Plot if <20 clusters
            if len(plotdf[cluster_column].unique()) < 20:
                dsshow(
                    plotdf,
                    ds.Point(f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"),
                    aggregator=ds.count_cat(cluster_column),
                    aspect="auto",
                    ax=axs[i // grid_size, i % grid_size],
                )
                axs[i // grid_size, i % grid_size].set_title(cluster_column)
        plt.savefig(f"{self.configuration.output}_umap-n{n_neighbors}_trajectories_clustered.png")
        plt.close()

    def run_kmeans(self, data: np.ndarray, n_clusters: int = 100) -> None:
        """Run KMeans on data array."""
        logger.info("Running KMeans clustering with %d clusters", n_clusters)
        kmeans = KMeans(n_clusters=n_clusters, n_init=10, max_iter=1000, random_state=42)
        kmeans.fit(data)
        column_name = f"kmeans_{n_clusters}"
        self.add_column(column_name, kmeans.labels_)
        self.cluster_columns.append(column_name)

    def run_leiden(self, data: np.ndarray, n_neighbors: int = 100, preprocess: bool = False) -> None:
        """Perform Leiden clustering on an array.

        Arguments:
        ---------
        data: np.ndarray
            Array of embeddings
        n_neighbors: int
            Number of neighbors to use for KNN graph
        preprocess: bool
            Whether to preprocess the data with StandardScaler

        """
        # Preprocessing: Standardize features by removing the mean and scaling to unit variance
        if preprocess:
            scaler = StandardScaler()
            data_scaled = scaler.fit_transform(data)
        else:
            data_scaled = data

        # Construct a k-nearest neighbors graph
        knn_graph = kneighbors_graph(data_scaled, n_neighbors=n_neighbors, mode="connectivity", include_self=True)

        # Convert to igraph
        sources, targets = knn_graph.nonzero()

        # Ensure the knn_graph is in a compatible sparse format
        if not isinstance(knn_graph, np.ndarray | np.generic):
            knn_graph = knn_graph.tocsr()

        # Extract weights directly from the sparse matrix
        weights = knn_graph[sources, targets].data

        # Flatten the weights to a 1D list if they are not already
        weights = np.array(weights).flatten().astype(float)

        # Create igraph Graph
        g = ig.Graph(
            n=data_scaled.shape[0],
            edges=list(zip(sources.tolist(), targets.tolist())),
            edge_attrs={"weight": weights.tolist()},
            directed=False,
        )

        # Perform Leiden clustering
        partition = leidenalg.find_partition(g, leidenalg.ModularityVertexPartition, weights="weight")
        clusters = partition.membership

        # Save results
        self.add_column("leiden", clusters)
        self.cluster_columns.append("leiden")



import numpy as np
import pandas as pd
import pytest

from jointly_hic.core.config import TrajectoryAnalysisConfig
from jointly_hic.core.trajectory_analyzer import TrajectoryAnalyzer


@pytest.fixture
def config():
    return TrajectoryAnalysisConfig(
        parquet_file="test.parquet",
        output="test_output",
        kmeans_clusters=[2, 3],
        leiden_neighbors=10,
        umap_neighbors=[10, 15],
    )


@pytest.fixture
def embeddings():
    return pd.DataFrame(
        {
            "chrom": ["chr1"] * 1000,
            "start": list(range(1, 501)) + list(range(1, 501)),
            "end": list(range(2, 502)) + list(range(2, 502)),
            "PCA1": np.random.rand(1000),
            "PCA2": np.random.rand(1000),
            "PCA3": np.random.rand(1000),
            "PCA4": np.random.rand(1000),
            "PCA5": np.random.rand(1000),
            "PCA6": np.random.rand(1000),
            "good_bin": np.random.choice([True, False], size=1000, p=[0.95, 0.05]),
            "filename": ["file1"] * 500 + ["file2"] * 500,
        }
    )


@pytest.fixture
def trajectory_analyzer(config, embeddings):
    return TrajectoryAnalyzer(config, embeddings)


def test_init_no_embeddings(mocker, config):
    # Patch pd.read_parquet
    mock_read_parquet = mocker.patch("pandas.read_parquet")
    analyzer = TrajectoryAnalyzer(config)
    assert analyzer.configuration == config
    mock_read_parquet.assert_called_once_with(config.parquet_file)
    assert analyzer.trajectory_df == mock_read_parquet.return_value.pivot.return_value


def test_init_with_embeddings(mocker, config, embeddings):
    # Patch pd.read_parquet
    mock_read_parquet = mocker.patch("pandas.read_parquet")
    analyzer = TrajectoryAnalyzer(config, embeddings)
    assert analyzer.configuration == config
    mock_read_parquet.assert_not_called()
    assert analyzer.trajectory_df.shape == (500, 16)
    assert "PCA1_file1" in analyzer.trajectory_df.columns
    assert "PCA1_file2" in analyzer.trajectory_df.columns
    assert "PCA6_file1" in analyzer.trajectory_df.columns
    assert "PCA6_file2" in analyzer.trajectory_df.columns
    assert "chrom" in analyzer.trajectory_df.columns
    assert "start" in analyzer.trajectory_df.columns
    assert "end" in analyzer.trajectory_df.columns
    assert "good_bin" in analyzer.trajectory_df.columns


def test_run(mocker, trajectory_analyzer):
    # Patch run_kmeans, run_leiden, run_umap, to_parquet, to_csv
    mock_run_kmeans = mocker.patch("jointly_hic.core.trajectory_analyzer.TrajectoryAnalyzer.run_kmeans")
    mock_run_leiden = mocker.patch("jointly_hic.core.trajectory_analyzer.TrajectoryAnalyzer.run_leiden")
    mock_run_umap = mocker.patch("jointly_hic.core.trajectory_analyzer.TrajectoryAnalyzer.run_umap")
    mock_to_parquet = mocker.patch("pandas.DataFrame.to_parquet")
    mock_to_csv = mocker.patch("pandas.DataFrame.to_csv")
    trajectory_analyzer.run()
    assert mock_run_kmeans.call_count == 2
    mock_run_leiden.assert_called_once()
    assert mock_run_umap.call_count == 3
    mock_to_parquet.assert_called_once_with(f"{trajectory_analyzer.configuration.output}_trajectories.pq", index=False)
    mock_to_csv.assert_called_once_with(f"{trajectory_analyzer.configuration.output}_trajectories.csv.gz", index=False)


def test_run_umap(mocker, trajectory_analyzer):
    # Patch plt.subplots
    mock_subplots = mocker.patch("matplotlib.pyplot.subplots", return_value=(None, None))
    # Patch savefig and close
    mock_savefig = mocker.patch("matplotlib.pyplot.savefig")
    mock_close = mocker.patch("matplotlib.pyplot.close")
    trajectory_analyzer.run_umap(trajectory_analyzer.data_values, n_neighbors=10)
    assert "umap1_n10" in trajectory_analyzer.trajectory_df.columns
    assert "umap2_n10" in trajectory_analyzer.trajectory_df.columns
    mock_subplots.assert_called_once()
    mock_savefig.assert_called_once_with(
        f"{trajectory_analyzer.configuration.output}_umap-n10_trajectories_clustered.png"
    )
    mock_close.assert_called_once()


def test_run_kmeans(mocker, trajectory_analyzer):
    trajectory_analyzer.run_kmeans(trajectory_analyzer.data_values, n_clusters=10)
    assert "kmeans_10" in trajectory_analyzer.trajectory_df.columns


@pytest.mark.parametrize("preprocess", [True, False])
def test_run_leiden(trajectory_analyzer, preprocess):
    trajectory_analyzer.run_leiden(trajectory_analyzer.data_values, n_neighbors=10, preprocess=preprocess)
    assert "leiden" in trajectory_analyzer.trajectory_df.columns



"""Create an HDF5 database of metadata, embeddings and tracks from jointly."""

import argparse
import concurrent
import logging
import threading
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor

import h5py
import numpy as np
import pandas as pd
import yaml

from jointly_hic.notebook_utils.encode_utils import EncodeFile, get_signal

# HDF5 File Layout:
# Groups:
# bins/: Contains datasets for bin information (chrom, start, end).
#     /chrom
#     /start
#     /end
#     /good_bin
#     /bin_name
# embeddings/: Contains datasets for PCA, UMAP, and other embeddings.
#     /PCA1, /PCA2, /PCA3, ...
#     /umap1_n500, /umap2_n500, ...
#     /leiden1, /leiden2, ...
#     /kmeans1, /kmeans2, ...
# metadata/: Contains metadata for experiments, tracks, and other annotations.
#     /experiment
#     /track_metadata
# tracks/: Contains datasets for tracks from bigwig files.
#     /ENCFF0001, /ENCFF0002, ...


# Lock
lock = threading.Lock()

# Set up logging
logger = logging.getLogger("joint_pca")


def create_hdf5(output_file: str) -> None:
    """Create an empty HDF5 file with the required structure.

    Args:
    ----
        output_file (str): Path to the output HDF5 file.

    """
    with h5py.File(output_file, "w") as hdf:
        hdf.create_group("bins")
        hdf.create_group("embeddings")
        hdf.create_group("metadata")
        hdf.create_group("tracks")
    logger.info(f"HDF5 file created at {output_file}")


def load_bins_from_embeddings(embeddings_file: str) -> pd.DataFrame:
    """Load bin information from the embeddings file.

    Args:
    ----
        embeddings_file (str): Path to the parquet file containing embeddings.
        accession_column (str): Column name used as the accession identifier.

    Returns:
    -------
        pd.DataFrame: DataFrame containing bin information.

    """
    df = pd.read_parquet(embeddings_file)
    bins = df[["chrom", "start", "end", "good_bin"]].drop_duplicates().sort_values(["chrom", "start", "end"])
    return bins


def store_bins(hdf5_file: str, parquet_file: str) -> None:
    """Store bin information in the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        parquet_file (str): Path to the parquet file containing embeddings.
        accession_column (str): Column name used as the accession identifier.

    """
    df = load_bins_from_embeddings(parquet_file)

    with h5py.File(hdf5_file, "a") as hdf:
        bins_group = hdf.require_group("bins")
        bins_group.create_dataset("chrom", data=np.array(df["chrom"], dtype="S"))
        bins_group.create_dataset("start", data=df["start"].values.astype(np.int64))
        bins_group.create_dataset("end", data=df["end"].values.astype(np.int64))
        bins_group.create_dataset("good_bin", data=df["good_bin"].values.astype(np.bool_))
        bin_names = [f"{chrom}:{start}-{end}" for chrom, start, end in zip(df["chrom"], df["start"], df["end"])]
        bins_group.create_dataset("bin_name", data=np.array(bin_names, dtype="S"))
    logger.info("Bin information stored successfully.")


def load_bins(hdf5_file: str) -> pd.DataFrame:
    """Load bin information from the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.

    Returns:
    -------
        pd.DataFrame: DataFrame containing bin information.

    """
    with h5py.File(hdf5_file, "r") as hdf:
        bins = hdf["bins"]
        bin_df = pd.DataFrame(
            {
                "chrom": [chrom.decode("utf-8") for chrom in bins["chrom"]],
                "start": bins["start"][:],
                "end": bins["end"][:],
                "good_bin": bins["good_bin"][:],
                "bin_name": [name.decode("utf-8") for name in bins["bin_name"]],
            }
        )

    return bin_df


def store_experiment_metadata(hdf5_file: str, yaml_file: str) -> None:
    """Store experiment metadata in the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        yaml_file (str): Path to the YAML file containing experiment metadata.

    """
    with open(yaml_file) as file:
        metadata_list = yaml.safe_load(file)

    metadata_df = pd.DataFrame(metadata_list)

    # Convert all columns to string to avoid object dtype issues
    metadata_df = metadata_df.map(str)

    # Convert DataFrame to records
    metadata_records = metadata_df.to_records(index=False)

    # Define dtype explicitly to avoid issues with object dtype
    dtype = np.dtype([(col, h5py.special_dtype(vlen=str)) for col in metadata_df.columns])

    with h5py.File(hdf5_file, "a") as hdf:
        metadata_group = hdf.require_group("metadata")
        if "experiment_metadata" in metadata_group:
            del metadata_group["experiment_metadata"]
        metadata_group.create_dataset("experiment_metadata", data=metadata_records, dtype=dtype)
    logger.info("Experiment metadata stored successfully.")


def load_experiment_metadata(hdf5_file: str) -> pd.DataFrame:
    """Load experiment metadata from the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.

    Returns:
    -------
        pd.DataFrame: DataFrame containing experiment metadata.

    """
    with h5py.File(hdf5_file, "r") as hdf:
        metadata_group = hdf["metadata"]
        metadata = metadata_group["experiment_metadata"]

        # Create a DataFrame and decode byte arrays to strings
        metadata_df = pd.DataFrame(metadata[:])
        metadata_df = metadata_df.map(lambda x: x.decode("utf-8"))

    return metadata_df


def store_embeddings(hdf5_file: str, parquet_file: str, accession_column: str) -> None:
    """Store embeddings in the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        parquet_file (str): Path to the parquet file containing embeddings.
        accession_column (str): Column name used as the accession identifier.

    """
    df = pd.read_parquet(parquet_file)

    # Load bins
    bins_df = load_bins(hdf5_file)

    # Load experiment metadata to determine the order of experiments
    experiment_metadata = load_experiment_metadata(hdf5_file)
    accessions = experiment_metadata["accession"]

    # Get columns with embeddings
    embedding_columns = [
        c for c in df.columns if "PCA" in c or "NMF" in c or "SVD" in c or "umap" in c or "leiden" in c or "kmeans" in c
    ]

    # Store the embeddings in the HDF5 file
    with h5py.File(hdf5_file, "a") as hdf:
        embeddings_group = hdf.require_group("embeddings")
        for i, embedding in enumerate(embedding_columns):
            logger.info(f"[{i + 1}/{len(embedding_columns)}] Storing embeddings for {embedding}.")
            # Create an empty matrix with rows as bins and columns as experiments
            num_bins = bins_df.shape[0]
            num_experiments = len(accessions)
            embedding_matrix = np.full((num_bins, num_experiments), np.nan, dtype=np.float64)

            # Fill the matrix according to the order in the metadata table
            for i, accession in enumerate(accessions):
                group = df.loc[df[accession_column] == accession]
                # Add `bin_name` to the group
                group.loc[:, "bin_name"] = [
                    f"{chrom}:{start}-{end}" for chrom, start, end in zip(group["chrom"], group["start"], group["end"])
                ]
                # Reorder embedding rows based on bins_df using pandas indexes
                embedding_matrix[:, i] = group.set_index("bin_name").loc[bins_df["bin_name"], embedding].values

            # Store the matrix in the HDF5 file
            embeddings_group.create_dataset(embedding, data=embedding_matrix)


def load_embeddings(
    hdf5_file: str,
    embeddings: list[str],
    experiment_metadata: pd.DataFrame,
    bins: pd.DataFrame,
    drop_bad_bins: bool = False,
) -> pd.DataFrame:
    """Load and pivot embeddings into a `tall` DataFrame, horizontally merged with bins and experiment metadata.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        embeddings (List[str]): List of embeddings to load.
        experiment_metadata (pd.DataFrame): DataFrame containing experiment metadata.
        bins (pd.DataFrame): DataFrame containing bin information.
        drop_bad_bins (bool): Whether to drop bad bins.

    Returns:
    -------
        pd.DataFrame: DataFrame containing combined embeddings, bins, and experiment metadata.

    """
    with h5py.File(hdf5_file, "r") as hdf:
        embeddings_group = hdf["embeddings"]

        # List to store data frames
        dfs = []

        for embedding in embeddings:
            if embedding in embeddings_group:
                # Load the embedding matrix
                embedding_matrix = embeddings_group[embedding][()]

                # Ensure that the number of columns matches the number of experiments
                if embedding_matrix.shape[1] != len(experiment_metadata):
                    raise ValueError("Mismatch between embedding and experiment metadata")

                # Ensure the bins match the number of rows
                if embedding_matrix.shape[0] != bins.shape[0]:
                    raise ValueError("Mismatch between embedding and bin information")

                # Create a DataFrame for the current embedding
                embedding_df = pd.DataFrame(embedding_matrix, columns=experiment_metadata["accession"])

                # Add bin information by horizontally concatenating the DataFrames, ignoring the index
                embedding_df = pd.concat([bins, embedding_df], axis=1)

                # Melt the DataFrame to a tall format
                embedding_df = embedding_df.melt(
                    id_vars=["bin_name", "chrom", "start", "end", "good_bin"],
                    var_name="accession",
                    value_name=embedding,
                ).reset_index()

                # Store the DataFrame in the list
                dfs.append(embedding_df)

    # Combine dataframes
    combined_df = dfs[0]
    for i, embedding in enumerate(embeddings):
        combined_df[embedding] = dfs[i][embedding]

    # Drop bad bins if required
    if drop_bad_bins:
        combined_df = combined_df[combined_df["good_bin"]]

    return combined_df.sort_values(["accession", "chrom", "start"])


def add_bigwig_track(
    hdf5_file: str, signal: pd.DataFrame, accession: str, assay: str, experiment: str, biosample: str
) -> None:
    """Add a bigwig track to the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        signal (pd.DataFrame): DataFrame containing the signal.
        accession (str): The accession identifier for the track.
        assay (str): The assay name.
        experiment (str): The experiment name.
        biosample (str): The biosample name.

    """
    with h5py.File(hdf5_file, "a") as hdf:
        # Add track data
        tracks_group = hdf.require_group("tracks")
        tracks_group.create_dataset(accession, data=np.array(signal, dtype=np.float64))

        # Add track metadata
        metadata_group = hdf.require_group("metadata/track_metadata")
        track_group = metadata_group.require_group(accession)
        track_group.attrs["assay"] = assay
        track_group.attrs["experiment"] = experiment
        track_group.attrs["biosample"] = biosample


def add_encode_tracks_to_hdf5(track_metadata_df: pd.DataFrame, hdf5_file: str) -> None:
    """Add tracks and their metadata to the HDF5 file.

    Args:
    ----
        track_metadata_df (pd.DataFrame): DataFrame containing track metadata.
        hdf5_file (str): Path to the HDF5 file.

    """
    number_of_tracks = track_metadata_df.shape[0]
    logger.info(f"Adding {number_of_tracks} tracks to the HDF5 file {hdf5_file}")

    # Load bin information from the HDF5 file
    bins_df = load_bins(hdf5_file)

    def process_entry(entry):
        """Process a single track entry."""
        acc = entry["accession"]
        logger.info(f"Loading signal for track: {acc}")
        with EncodeFile(acc, "bw") as bigwig:
            bw_signal = get_signal(bigwig, bins_df)
            return bw_signal, acc, entry["assay"], entry["experiment"], entry["biosample"]

    # Process each track entry in parallel
    with ThreadPoolExecutor(8) as executor:
        futures = [executor.submit(process_entry, entry) for _, entry in track_metadata_df.iterrows()]

        for i, future in enumerate(concurrent.futures.as_completed(futures)):
            try:
                signal, accession, assay, experiment, biosample = future.result()
                logger.info(f"[{i + 1}/{number_of_tracks}] Adding track {accession} to HDF5 file.")
                with lock:
                    add_bigwig_track(hdf5_file, signal, accession, assay, experiment, biosample)
            except Exception as e:
                logger.error(f"Error processing track: {e}")

    logger.info("All tracks and metadata added successfully.")


def load_track_metadata(hdf5_file: str) -> pd.DataFrame:
    """Load track metadata from the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.

    Returns:
    -------
        pd.DataFrame: DataFrame containing track metadata.

    """
    with h5py.File(hdf5_file, "r") as hdf:
        metadata_group = hdf["metadata/track_metadata"]
        track_metadata = []
        for accession in metadata_group:
            track_group = metadata_group[accession]
            track_metadata.append(
                {
                    "accession": accession,
                    "assay": track_group.attrs["assay"],
                    "experiment": track_group.attrs["experiment"],
                    "biosample": track_group.attrs["biosample"],
                }
            )
    return pd.DataFrame(track_metadata)


def load_tracks(hdf5_file: str, accessions: list[str], bins: pd.DataFrame, drop_bad_bins: bool = False) -> pd.DataFrame:
    """Load track signal from the HDF5 file.

    Args:
    ----
        hdf5_file (str): Path to the HDF5 file.
        accessions (List[str]): The accession identifiers for the tracks.
        bins (pd.DataFrame): DataFrame containing bin information.
        drop_bad_bins (bool): Whether to drop bad bins.

    Returns:
    -------
        pd.DataFrame: DataFrame containing the track signal.

    """
    tracks = bins.copy()

    with h5py.File(hdf5_file, "r") as hdf:
        tracks_group = hdf["tracks"]
        for i, accession in enumerate(accessions):
            signal = tracks_group[accession][()]
            tracks[accession] = signal
            if i % 5 == 0:
                tracks = tracks.copy()

        # Drop bad bins if required
        if drop_bad_bins:
            tracks = tracks[tracks["good_bin"]]

        return tracks


class JointDb:
    """Class to query the HDF5 database."""

    def __init__(self, hdf5_file: str):
        """Initialize the JointDb object."""
        self.hdf5_file = hdf5_file
        self._bins = None
        self._experiment_metadata = None
        self._track_metadata = None

    @property
    def bins(self) -> pd.DataFrame:
        """Load bin information from the HDF5 file.

        Returns
        -------
            pd.DataFrame: DataFrame containing bin information.

        """
        if self._bins is None:
            self._bins = load_bins(self.hdf5_file)
        return self._bins

    @property
    def experiment_metadata(self) -> pd.DataFrame:
        """Load experiment metadata from the HDF5 file.

        Returns
        -------
            pd.DataFrame: DataFrame containing experiment metadata.

        """
        if self._experiment_metadata is None:
            self._experiment_metadata = load_experiment_metadata(self.hdf5_file)
        return self._experiment_metadata

    @property
    def track_metadata(self) -> pd.DataFrame:
        """Load track metadata from the HDF5 file.

        Returns
        -------
            pd.DataFrame: DataFrame containing track metadata.

        """
        if self._track_metadata is None:
            self._track_metadata = load_track_metadata(self.hdf5_file)
        return self._track_metadata

    def get_embeddings(
        self, embeddings: list[str], accessions: list[str] | None = None, drop_bad_bins: bool = False
    ) -> pd.DataFrame:
        """Load embeddings from the HDF5 file.

        Args:
        ----
            embeddings (List[str]): List of embeddings to load.
            accessions (List[str]): The accession identifiers for the embeddings.
            drop_bad_bins (bool): Whether to drop bad bins.

        Returns:
        -------
            pd.DataFrame: DataFrame containing embeddings.

        """
        embeddings = load_embeddings(self.hdf5_file, embeddings, self.experiment_metadata, self.bins, drop_bad_bins)

        # Filter embeddings by accessions if required
        if accessions is not None:
            embeddings = embeddings[embeddings["accession"].isin(accessions)]

        return embeddings

    def get_tracks(self, accessions: list[str], drop_bad_bins: bool = False) -> pd.DataFrame:
        """Load track signal from the HDF5 file.

        Args:
        ----
            accessions (List[str]): The accession identifiers for the tracks.
            drop_bad_bins (bool): Whether to drop bad bins.

        Returns:
        -------
            pd.DataFrame: DataFrame containing the track signal.

        """
        return load_tracks(self.hdf5_file, accessions, self.bins, drop_bad_bins)

    def get_track_overlay(
        self,
        assay: str,
        embeddings: list[str],
        accessions: list[str] | None = None,
        drop_bad_bins: bool = False,
        scaler: Callable | None = None,
        biosample_mapper: Callable | None = None,
    ) -> pd.DataFrame:
        """Load embeddings & merge tracks on biosample, for a specific assay.

        Args:
        ----
            assay (str): The assay type.
            embeddings (List[str]): List of embeddings to load.
            accessions (List[str]): The embedding accessions to load.
            drop_bad_bins (bool): Whether to drop bad bins.
            scaler (Callable): A function to scale the signal.
            biosample_mapper (Callable): A function to map biosample names.

        Returns:
        -------
            pd.DataFrame: DataFrame containing embeddings, bins, and experiment metadata.

        """
        # Load embeddings
        embeddings_df = self.get_embeddings(embeddings, accessions=accessions, drop_bad_bins=drop_bad_bins)
        embeddings_df = pd.merge(self.experiment_metadata, embeddings_df, on="accession", how="inner")
        embeddings_df = embeddings_df[["bin_name", "biosample", "accession", *embeddings]]

        # If a biosample mapper is provided, apply it
        if biosample_mapper is not None:
            embeddings_df["biosample"] = embeddings_df["biosample"].map(biosample_mapper)
            # Example biosample mapper: lambda x: x.split("-")[0]

        # Load tracks for the given assay
        assay_tracks = self.track_metadata[self.track_metadata["assay"] == assay]
        track_df = self.get_tracks(assay_tracks["accession"].tolist(), drop_bad_bins)

        # Melt track_df and merge with track_metadata
        track_df = track_df.melt(
            id_vars=["bin_name", "chrom", "start", "end", "good_bin"], var_name="bw_accession", value_name="signal"
        ).reset_index(drop=True)
        track_df = pd.merge(track_df, self.track_metadata, left_on="bw_accession", right_on="accession", how="inner")

        # Scale the signal if a scaler is provided
        if scaler is not None:
            track_df = scaler(track_df)

        track_df = track_df[["bin_name", "biosample", "signal"]]

        # Merge with embeddings, then groupby and average signal so each sample-bin has a single signal
        result_df = (
            embeddings_df.merge(track_df, on=["bin_name", "biosample"], how="inner")
            .groupby(["bin_name", "biosample", "accession"])
            .mean()
            .reset_index()
        )
        return result_df

    def __repr__(self):
        """Return a string representation of the JointDb object."""
        return (
            f"JointDb(file='{self.hdf5_file}', bins={self.bins.shape[0]}, "
            f"experiments={self.experiment_metadata.shape[0]}, tracks={self.track_metadata.shape[0]})"
        )


def main(
    experiments_file: str, tracks_file: str | None, embeddings_file: str, accession_column: str, output_file: str
) -> None:
    """Create the HDF5 file with bins, metadata, embeddings, and tracks.

    Args:
    ----
        experiments_file (str): Path to the YAML file containing experiment metadata.
        tracks_file (str): Path to the YAML file containing track metadata.
        embeddings_file (str): Path to the parquet file containing embeddings.
        accession_column (str): Column name used as the accession identifier.
        output_file (str): Path to the output HDF5 file.

    """
    # Step 1: Create the HDF5 file
    logger.info(f"Creating HDF5 file at {output_file}")
    create_hdf5(output_file)

    # Step 2: Load and store bins (assuming bins are part of the embeddings file)
    logger.info("Loading and storing bins...")
    store_bins(output_file, embeddings_file)

    # Step 3: Load and store experiment metadata
    logger.info("Loading and storing experiment metadata...")
    store_experiment_metadata(output_file, experiments_file)

    # Step 4: Load and store embeddings
    logger.info("Loading and storing embeddings...")
    store_embeddings(output_file, embeddings_file, accession_column)

    # Step 5: Load and store track metadata and tracks
    if tracks_file is not None:
        logger.info("Loading and storing track metadata and tracks...")
        with open(tracks_file) as file:
            track_metadata_list = yaml.safe_load(file)

        track_metadata_df = pd.DataFrame(track_metadata_list)
        add_encode_tracks_to_hdf5(track_metadata_df, output_file)

    logger.info(f"HDF5 file '{output_file}' created successfully with all data.")


def run_main(arguments):
    """Run the main function with arguments."""
    main(arguments.experiments, arguments.tracks, arguments.embeddings, arguments.accession, arguments.output)


def configure_hdf5db_parser(parser):
    """Configure the HDF5 database parser."""
    parser.add_argument("--experiments", required=True, help="Path to the YAML file containing experiment metadata.")
    parser.add_argument("--tracks", required=False, help="Path to the YAML file containing track metadata.")
    parser.add_argument("--embeddings", required=True, help="Path to the parquet file containing embeddings.")
    parser.add_argument("--accession", required=True, help="Column name used as the accession identifier.")
    parser.add_argument("--output", required=True, help="Path to the output HDF5 file.")
    parser.set_defaults(func=run_main)


if __name__ == "__main__":
    # Define the CLI arguments
    local_parser = argparse.ArgumentParser(
        description="Create an HDF5 file with bins, metadata, embeddings, and tracks."
    )
    configure_hdf5db_parser(local_parser)

    # Parse the arguments
    args = local_parser.parse_args()

    # Call the main function with the parsed arguments
    main(args.experiments, args.tracks, args.embeddings, args.accession, args.output)



import tempfile

import pandas as pd
import pytest

import jointly_hic.notebook_utils.encode_utils
from jointly_hic.notebook_utils.encode_utils import get_bigwigs_from_experiment_report, EncodeFile, get_signal


@pytest.fixture
def bins():
    return pd.DataFrame(
        {
            "chrom": ["chr1", "chr1", "chr1"],
            "start": [0, 100, 200],
            "end": [100, 200, 300],
        }
    )


def test_encode_utils(bins, tmpdir):
    bigwigs = get_bigwigs_from_experiment_report("liver")
    assert len(bigwigs) > 0

    # Create a new temporary directory for testing the cache
    with tempfile.TemporaryDirectory(dir="./") as cache:
        jointly_hic.notebook_utils.encode_utils.CACHE_DIR = cache
        with EncodeFile(list(bigwigs.values())[0], "bw") as bigwig:
            signal = get_signal(bigwig, bins)
            assert len(signal) == len(bins)



"""Main entrypoint for the jointly_hic package."""

import logging

from jointly_hic.parser import JointlyCommandLineInterface


def setup_logging(args):
    """Set up logging to file and console."""
    logger = logging.getLogger("joint_pca")
    logger.handlers.clear()
    logger.setLevel(args.log_level)

    # Setup logging to file
    handler = logging.FileHandler(args.log, mode="w", encoding="utf-8")
    handler.setLevel(args.log_level)
    formatter = logging.Formatter("%(asctime)s::%(levelname)s::%(funcName)s:   %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Setup logging to console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(args.log_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # Disable propagation to parent loggers
    logger.propagate = False
    return logger


def main():
    """Run the jointly_hic program."""
    # Parse arguments as dispatch subcommand
    parser = JointlyCommandLineInterface().parser
    args = parser.parse_args()
    logger = setup_logging(args)
    try:
        logger.info("Starting jointly-hic")
        args.func(args)
        logger.info("Finished jointly-hic")
    except Exception as e:
        logger.exception("Error running jointly-hic")
        logger.exception(e)
        raise e



import numpy as np
import pandas as pd
import pytest

from jointly_hic.notebook_utils.bin_operations import make_chromarms, distance_traveled_by_bin


def test_distance_traveled_by_bin():
    # Mock input DataFrame
    df = pd.DataFrame(
        {
            "chrom": ["chr1", "chr1", "chr1"],
            "start": [0, 100, 200],
            "PCA1_T0": [1, 2, 3],
            "PCA2_T0": [4, 5, 6],
            "PCA1_T1": [1, 2, 4],
            "PCA2_T1": [4, 5, 7],
            "Time": ["T0", "T0", "T0"],
        }
    )
    time_column = "Time"
    time_points = ["T0"]
    prefix = "PCA"

    result_df = distance_traveled_by_bin(df, time_column, time_points, prefix)

    # Expected results
    expected_df = pd.DataFrame(
        {
            "chrom": ["chr1", "chr1", "chr1"],
            "start": [0, 100, 200],
            "total_distance": [0.0, 0.0, 0.0],
        }
    )

    pd.testing.assert_frame_equal(result_df, expected_df)


def test_make_chromarms():
    chromsizes = pd.Series({"chr1": 1000, "chr2": 2000})
    mids = {"chr1": 500, "chr2": 1000}
    binsize = 100

    result_df = make_chromarms(chromsizes, mids, binsize)

    expected_df = pd.DataFrame(
        {
            "chrom": ["chr1", "chr1", "chr2", "chr2"],
            "start": [0, 500, 0, 1000],
            "end": [500, 1000, 1000, 2000],
            "name": ["chr1p", "chr1q", "chr2p", "chr2q"],
        }
    )

    pd.testing.assert_frame_equal(result_df, expected_df)



"""Class to handle post-processing of embeddings."""

import logging

import datashader as ds
import igraph as ig
import leidenalg
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import umap
import umap.plot
from datashader.mpl_ext import dsshow
from sklearn.cluster import KMeans
from sklearn.neighbors import kneighbors_graph
from sklearn.preprocessing import StandardScaler

from jointly_hic.core.config import PostProcessingConfig

logger = logging.getLogger("joint_pca")
MAX_PLOT_CLUSTERS = 20


class PostProcessor:
    """Class to handle post-processing of embeddings.

    PostProcessor is the main interface to handle basic Jointly-HiC post-processing. It takes a PostProcessingConfig
    and a pandas DataFrame of embeddings as input. It then runs KMeans clustering and UMAP on the embeddings and
    saves the results to parquet and csv files. It also creates basic plots of the embeddings.

    Configure the PostProcessor with a PostProcessingConfig.
    """

    def __init__(self, configuration: PostProcessingConfig, embeddings: pd.DataFrame | None = None):
        """Initialize PostProcessor."""
        self.configuration: PostProcessingConfig = configuration
        logger.info("PostProcessor :: configuration: \n\n%s", configuration)
        self.cluster_columns: list[str] = []
        if embeddings is None:
            logger.info("Loading embeddings from parquet file: %s", self.configuration.parquet_file)
            loaded_embeddings = pd.read_parquet(self.configuration.parquet_file)
        else:
            loaded_embeddings = embeddings.copy()

        # Ensure types
        loaded_embeddings["chrom"] = loaded_embeddings["chrom"].astype(str)
        loaded_embeddings["start"] = loaded_embeddings["start"].astype(int)
        loaded_embeddings["end"] = loaded_embeddings["end"].astype(int)
        self.embeddings: pd.DataFrame = loaded_embeddings

    @property
    def components(self) -> np.ndarray:
        """Return columns that start with 'PC#' and good_bin is True as numpy array."""
        try:
            return self.embeddings.loc[
                self.embeddings.good_bin,
                [col for col in self.embeddings.columns if col.startswith(self.configuration.method)],
            ].values
        except KeyError:
            logger.warning("No columns starting with %s found. Trying: PCA, NMF, SVD.", self.configuration.method)
            try:
                return self.embeddings.loc[
                    self.embeddings.good_bin, [col for col in self.embeddings.columns if col.startswith("PCA")]
                ].values
            except KeyError:
                logger.warning("No columns starting with PCA found. Trying: NMF, SVD.")
            try:
                return self.embeddings.loc[
                    self.embeddings.good_bin, [col for col in self.embeddings.columns if col.startswith("NMF")]
                ].values
            except KeyError:
                logger.warning("No columns starting with NMF found. Trying: SVD.")
            try:
                return self.embeddings.loc[
                    self.embeddings.good_bin, [col for col in self.embeddings.columns if col.startswith("SVD")]
                ].values
            except KeyError:
                logger.warning("No columns starting with SVD found. Trying: PC.")

        raise Exception("No columns starting with PCA, NMF, or SVD found.")

    def normalize_embeddings(self):
        """Normalize embeddings by dataset then rescale to overall norm."""
        # Compute overall norm
        norm = np.linalg.norm(self.components)

        # Group by 'filename' and rescale by group norm then overall norm
        for i, df in self.embeddings.groupby("filename"):
            logger.info(f"Normalizing embeddings for {i}")
            component_values = df.loc[
                df.good_bin, [col for col in df.columns if col.startswith(self.configuration.method)]
            ].values
            scaled_components = component_values / np.linalg.norm(component_values) * norm
            df.loc[df.good_bin, [col for col in df.columns if col.startswith(self.configuration.method)]] = (
                scaled_components
            )
            self.embeddings.update(df)
            logger.debug("Count of filenames: \n%s", self.embeddings.filename.value_counts())

    def add_column(self, column_name, values):
        """Add column to embeddings where good_bin is True and otherwise add NaN."""
        self.embeddings[column_name] = np.nan
        self.embeddings.loc[self.embeddings.good_bin, column_name] = values

    def run(self):
        """Run post-processing."""
        logger.info("Running post-processing")
        self.normalize_embeddings()

        # Set bin_name as chrom_start_end and compute bin distance variance
        self.embeddings["bin_name"] = (
            self.embeddings["chrom"]
            + "_"
            + self.embeddings["start"].astype(str)
            + "_"
            + self.embeddings["end"].astype(str)
        )
        self.embeddings["bin_variance"] = self.embeddings["bin_name"].map(self.compute_bin_distance_variance())

        components = self.components
        logger.info("Components shape: %s", components.shape)

        # Run kmeans clustering
        for n_clusters in self.configuration.kmeans_clusters:
            self.run_kmeans(components, n_clusters=n_clusters)

        # Run Leiden clustering
        for resolution in self.configuration.leiden_resolutions:
            self.run_leiden(components, resolution=resolution, n_neighbors=500)

        # Run UMAP and plot (requires clustering to be run first)
        for n_neighbors in self.configuration.umap_neighbours:
            self.run_umap(components, n_neighbors=n_neighbors)

        # Plot scores with clusters
        self.plot_scores()

        # Save embeddings to parquet and csv
        logger.info("Saving embeddings to parquet and csv")
        self.embeddings.to_csv(f"{self.configuration.output}_post_processed.csv.gz", index=False)
        self.embeddings.to_parquet(f"{self.configuration.output}_post_processed.pq", index=False)

        logger.info("Finished running post-processing")

    def plot_scores(self):
        """Plot the scores."""
        logger.info("Plotting scores")

        # Create plotdf from columns we want
        components = [f"{self.configuration.method}{i}" for i in range(1, 5)]
        plotdf = self.embeddings.loc[self.embeddings.good_bin, [*components, "good_bin", "filename"]]
        for cluster_column in self.cluster_columns:
            plotdf[cluster_column] = self.embeddings.loc[self.embeddings.good_bin, cluster_column].astype("category")

        # Truncate PC values at 1% and 99% percentile
        for pc in components:
            plotdf[pc] = plotdf[pc].clip(lower=plotdf[pc].quantile(0.005), upper=plotdf[pc].quantile(0.995))

        # Plot every pair of PCs using dsshow and a grid
        _, axs = plt.subplots(4, 4, figsize=(12, 12))
        for i, pc1 in enumerate(components):
            for j, pc2 in enumerate(components):
                if pc1 != pc2:
                    # Plot PC scores for pairs of PCs
                    dsshow(
                        plotdf,
                        ds.Point(pc1, pc2),
                        aspect="auto",
                        ax=axs[i, j],
                    )
                    axs[i, j].set_title(f"{pc1} vs {pc2}")
                if pc1 == pc2:
                    # Plot histogram of PC
                    axs[i, j].hist(plotdf[pc1], bins=100)
                    axs[i, j].set_title(pc1)
        plt.savefig(f"{self.configuration.output}_scores.png")
        plt.close()

        # Plot colored by cluster on a grid
        n_clusters = len(self.cluster_columns)
        grid_size = int(np.ceil(np.sqrt(n_clusters)))
        xmin = plotdf[components[0]].min()
        xmax = plotdf[components[0]].max()
        ymin = plotdf[components[1]].min()
        ymax = plotdf[components[1]].max()
        _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
        for i, cluster_column in enumerate(self.cluster_columns):
            # Plot if there are <= MAX_PLOT_CLUSTERS clusters
            if len(plotdf[cluster_column].unique()) <= MAX_PLOT_CLUSTERS:
                dsshow(
                    plotdf,
                    ds.Point(components[0], components[1]),
                    aggregator=ds.count_cat(cluster_column),
                    aspect="auto",
                    ax=axs[i // grid_size, i % grid_size],
                    x_range=(xmin, xmax),
                    y_range=(ymin, ymax),
                )
                axs[i // grid_size, i % grid_size].set_title(cluster_column)
        plt.savefig(f"{self.configuration.output}_scores_clustered.png")
        plt.close()

        # Plot PC1 vs PC2 colored by cluster for every filename
        n_files = len(plotdf.filename.unique())
        grid_size = int(np.ceil(np.sqrt(n_files)))
        _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
        for i, filename in enumerate(plotdf.filename.unique()):
            dsshow(
                plotdf[plotdf.filename == filename],
                ds.Point(components[0], components[1]),
                aspect="auto",
                ax=axs[i // grid_size, i % grid_size],
                x_range=(xmin, xmax),
                y_range=(ymin, ymax),
            )
            axs[i // grid_size, i % grid_size].set_title(filename)
        plt.savefig(f"{self.configuration.output}_scores_filenames.png")
        plt.close()

    def run_kmeans(self, components: np.ndarray, n_clusters: int = 8) -> None:
        """Run KMeans clustering on embeddings."""
        logger.info("Running KMeans clustering with %d clusters", n_clusters)
        # Initialize and fit the K-Means model
        kmeans = KMeans(n_clusters=n_clusters, n_init=10, max_iter=1000, random_state=42)
        kmeans.fit(components)
        column_name = f"kmeans_{n_clusters}"
        self.add_column(column_name, kmeans.labels_)
        self.cluster_columns.append(column_name)

    def run_leiden(
        self, components: np.ndarray, n_neighbors: int = 100, resolution: float = 1.0, preprocess: bool = False
    ) -> None:
        """Perform Leiden clustering on an array of embeddings.

        Arguments:
        ---------
        components: np.ndarray
            Array of embeddings
        n_neighbors: int
            Number of neighbors to use for KNN graph
        resolution: float
            Resolution parameter for Leiden clustering
        preprocess: bool
            Whether to preprocess the data with StandardScaler

        """
        logger.info("Running Leiden clustering with %d neighbors.", n_neighbors)
        # Preprocessing: Standardize features by removing the mean and scaling to unit variance
        if preprocess:
            scaler = StandardScaler()
            data_scaled = scaler.fit_transform(components)
        else:
            data_scaled = components

        # Construct a k-nearest neighbors graph
        knn_graph = kneighbors_graph(data_scaled, n_neighbors=n_neighbors, mode="connectivity", include_self=True)

        # Convert to igraph
        sources, targets = knn_graph.nonzero()

        # Ensure the knn_graph is in a compatible sparse format
        if not isinstance(knn_graph, np.ndarray | np.generic):
            knn_graph = knn_graph.tocsr()

        # Extract weights directly from the sparse matrix
        weights = knn_graph[sources, targets].data

        # Flatten the weights to a 1D list if they are not already
        weights = np.array(weights).flatten().astype(float)

        # Create igraph Graph
        g = ig.Graph(
            n=data_scaled.shape[0],
            edges=list(zip(sources.tolist(), targets.tolist())),
            edge_attrs={"weight": weights.tolist()},
            directed=False,
        )

        # Perform Leiden clustering
        partition = leidenalg.find_partition(
            g, leidenalg.RBERVertexPartition, weights="weight", resolution_parameter=resolution
        )
        clusters = partition.membership

        # Save results
        column_name = f"leiden_{str(resolution).replace('.', '_')}_n{n_neighbors}"
        self.add_column(column_name, clusters)
        self.cluster_columns.append(column_name)
        logger.info("Leiden clustering complete.")

    def run_umap(self, scores: np.ndarray, n_neighbors: int = 50) -> None:
        """Run UMAP on embeddings."""
        logger.info("Running UMAP with %d neighbors", n_neighbors)
        mapper = umap.UMAP(n_neighbors=n_neighbors)
        umap_embedding = mapper.fit_transform(scores)
        self.add_column(f"umap1_n{n_neighbors}", umap_embedding[:, 0])
        self.add_column(f"umap2_n{n_neighbors}", umap_embedding[:, 1])

        # Plot a umap for each clustering_column
        n_clusters = len(self.cluster_columns)
        grid_size = int(np.ceil(np.sqrt(n_clusters)))
        _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
        for i, cluster_column in enumerate(self.cluster_columns):
            plotdf = self.embeddings.loc[self.embeddings.good_bin, [f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"]]
            plotdf[cluster_column] = self.embeddings.loc[self.embeddings.good_bin, cluster_column].astype("category")
            # Only plot if <= MAX_PLOT_CLUSTERS clusters
            if len(plotdf[cluster_column].unique()) <= MAX_PLOT_CLUSTERS:
                dsshow(
                    plotdf,
                    ds.Point(f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"),
                    aggregator=ds.count_cat(cluster_column),
                    aspect="auto",
                    ax=axs[i // grid_size, i % grid_size],
                )
                axs[i // grid_size, i % grid_size].set_title(cluster_column)
        plt.savefig(f"{self.configuration.output}_umap-n{n_neighbors}_clustered.png")
        plt.close()

        # Plot a umap for each filename on a grid
        n_files = len(self.embeddings.filename.unique())
        grid_size = int(np.ceil(np.sqrt(n_files)))
        logger.debug("Plotting UMAP for each filename in:\n %s", self.embeddings.filename.value_counts())
        _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
        for i, filename in enumerate(self.embeddings.filename.unique()):
            plotdf = self.embeddings.loc[self.embeddings.filename == filename]
            plotdf = plotdf.loc[plotdf.good_bin, [f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"]]
            dsshow(
                plotdf,
                ds.Point(f"umap1_n{n_neighbors}", f"umap2_n{n_neighbors}"),
                aspect="auto",
                ax=axs[i // grid_size, i % grid_size],
            )
            axs[i // grid_size, i % grid_size].set_title(filename)
        plt.savefig(f"{self.configuration.output}_umap-n{n_neighbors}_filenames.png")
        plt.close()

    def compute_bin_distance_variance(self):
        """Compute the variance of a bin across all samples in component space.

        Group by bin_name, then compute the variance for each component across all samples, then sum the total variance.

        Returns
        -------
        bin_variance: dict
            A dictionary of {bin_name: variance} for each bin.

        """
        df = self.embeddings.loc[
            self.embeddings.good_bin,
            [col for col in self.embeddings.columns if col.startswith(self.configuration.method)] + ["bin_name"],
        ]
        bin_variance = df.groupby("bin_name").var().sum(axis=1)

        # Plot histogram of bin_variance
        plt.hist(bin_variance, bins=100)
        plt.title("Bin Distance Variance")
        plt.xlabel("Variance")
        plt.ylabel("Frequency")
        plt.savefig(f"{self.configuration.output}_bin_distance_variance.png")
        plt.close()

        return bin_variance.to_dict()



import os.path
from glob import glob
from uuid import uuid4

import bioframe
import numpy as np
import pytest

from jointly_hic.notebook_utils.heatmap import encode_chip_heatmap


@pytest.fixture
def bins():
    bins = bioframe.binnify(bioframe.fetch_chromsizes("hg38"), binsize=1000000)
    bins["feature1"] = np.random.rand(len(bins))
    bins["feature2"] = np.random.rand(len(bins))
    bins["name"] = "sample1"
    bins["cluster"] = np.random.choice([1, 2, 3], len(bins))
    return bins


def test_encode_chip_heatmap(bins):
    output_prefix = str(uuid4())
    encode_chip_heatmap(
        bins,
        data_cols=["feature1", "feature2"],
        name="sample1",
        bigwigs={},
        labels=["cluster"],
        output_prefix=output_prefix,
    )

    assert os.path.exists(f"{output_prefix}.pq")
    assert os.path.exists(f"{output_prefix}_heatmap.png")

    # Remove output files
    for f in glob(f"{output_prefix}*"):
        os.remove(f)



import pytest
from pydantic_core._pydantic_core import ValidationError

from jointly_hic.core.config import JointlyConfiguration, PostProcessingConfig, TrajectoryAnalysisConfig


def test_joint_pca_config():
    config = JointlyConfiguration(
        mcools=["test1.mcool", "test2.mcool"],
        output="test",
        resolution=10000,
        assembly="hg38",
        components=32,
        chrom_limit=23,
        method="PCA",
        percentile_top=99.5,
        percentile_bottom=1,
        batch_size=20000,
        log_level="INFO",
    )
    assert config.mcools == ["test1.mcool", "test2.mcool"]
    assert config.output == "test"
    assert config.resolution == 10000
    assert config.assembly == "hg38"
    assert config.components == 32
    assert config.chrom_limit == 23
    assert config.method == "PCA"
    assert config.percentile_top == 99.5
    assert config.percentile_bottom == 1
    assert config.batch_size == 20000
    assert config.log_level == "INFO"
    assert isinstance(str(config), str)

    # Test error for wrong types
    with pytest.raises(ValidationError):
        JointlyConfiguration(
            mcools="test1.mcool",
            output="test",
            resolution=10000,
            assembly="hg38",
            components=32,
            chrom_limit=23,
            method="PCA",
            percentile_top=99.5,
            percentile_bottom=1,
            batch_size=20000,
            log_level="INFO",
        )


def test_post_processing_config():
    config = PostProcessingConfig(
        parquet_file="test.parquet",
        output="test",
        umap_neighbours=[30, 100, 500],
        kmeans_clusters=[5, 10, 15, 20],
        log_level="INFO",
    )
    assert config.parquet_file == "test.parquet"
    assert config.output == "test"
    assert config.umap_neighbours == [30, 100, 500]
    assert config.kmeans_clusters == [5, 10, 15, 20]
    assert config.log_level == "INFO"
    assert isinstance(str(config), str)

    # Test error for wrong types
    with pytest.raises(ValidationError):
        PostProcessingConfig(
            parquet_file="test.parquet",
            output="test",
            umap_neighbours=30,
            kmeans_clusters=[5, 10, 15, 20, 25, 30],
            log_level="INFO",
        )


def test_trajectory_analysis_config():
    config = TrajectoryAnalysisConfig(
        parquet_file="test.parquet",
        output="test",
        umap_neighbours=[30, 100, 500],
        kmeans_clusters=[5, 10, 15, 20, 25, 30],
        log_level="INFO",
    )
    assert config.parquet_file == "test.parquet"
    assert config.output == "test"
    assert config.umap_neighbours == [30, 100, 500]
    assert config.kmeans_clusters == [5, 10, 15, 20, 25, 30]
    assert config.log_level == "INFO"
    assert isinstance(str(config), str)

    # Test error for wrong types
    with pytest.raises(ValidationError):
        TrajectoryAnalysisConfig(
            parquet_file="test.parquet",
            output="test",
            umap_neighbours=30,
            kmeans_clusters=[5, 10, 15, 20, 25, 30],
            log_level="INFO",
        )



from unittest.mock import patch, MagicMock

import pytest

from jointly_hic.entrypoint import setup_logging, main


def test_setup_logging(mocker):
    """Test setup_logging."""
    # Mock logging.getLogger
    get_logger_patch = mocker.patch("jointly_hic.entrypoint.logging.getLogger")
    # Mock logging.FileHandler
    file_handler_patch = mocker.patch("jointly_hic.entrypoint.logging.FileHandler")
    # Mock logging.StreamHandler
    mocker.patch("jointly_hic.entrypoint.logging.StreamHandler")
    # Mock logging.Formatter
    formatter_patch = mocker.patch("jointly_hic.entrypoint.logging.Formatter")

    # Create some CLI args
    args = mocker.Mock()
    args.log = "test_log"
    args.log_level = "INFO"

    # Test setup_logging
    setup_logging(args)

    # Test
    get_logger_patch.assert_called_once_with("joint_pca")
    get_logger_patch.return_value.handlers.clear.assert_called_once()
    get_logger_patch.return_value.setLevel.assert_called_once_with(args.log_level)
    file_handler_patch.assert_called_once_with(
        args.log,
        mode="w",
        encoding="utf-8",
    )
    file_handler_patch.return_value.setLevel.assert_called_once_with(args.log_level)
    formatter_patch.assert_called_once_with("%(asctime)s::%(levelname)s::%(funcName)s:   %(message)s")
    file_handler_patch.return_value.setFormatter.assert_called_once_with(formatter_patch.return_value)


def test_main_success():
    """Test the main function when it completes successfully."""
    with (
        patch("jointly_hic.entrypoint.JointlyCommandLineInterface") as mock_cli,
        patch("jointly_hic.entrypoint.setup_logging") as mock_setup_logging,
    ):
        # Mock the parser to simulate command line arguments being parsed successfully
        mock_parser = MagicMock()
        mock_cli.return_value.parser = mock_parser
        mock_parser.parse_args.return_value = MagicMock(func=lambda x: None)

        # Mock the logger to capture logging without sending output anywhere
        mock_logger = MagicMock()
        mock_setup_logging.return_value = mock_logger

        # Run the main function
        main()

        # Check that logging was called with the expected messages
        mock_logger.info.assert_any_call("Starting jointly-hic")
        mock_logger.info.assert_any_call("Finished jointly-hic")


def test_main_exception():
    """Test the main function when an exception is raised."""
    with (
        patch("jointly_hic.entrypoint.JointlyCommandLineInterface") as mock_cli,
        patch("jointly_hic.entrypoint.setup_logging") as mock_setup_logging,
    ):
        # Mock the parser to simulate command line arguments being parsed successfully
        mock_parser = MagicMock()
        mock_cli.return_value.parser = mock_parser
        error = Exception("Test error")
        mock_parser.parse_args.return_value = MagicMock(func=lambda x: (_ for _ in ()).throw(error))

        # Mock the logger to capture logging without sending output anywhere
        mock_logger = MagicMock()
        mock_setup_logging.return_value = mock_logger

        # Run the main function and expect it to raise an exception
        with pytest.raises(Exception) as exc_info:
            main()

        # Check that the exception was logged
        mock_logger.exception.assert_called_with(error)

        # Verify the exception message
        assert str(exc_info.value) == "Test error", "Unexpected exception message"
